//---------------------------------------------------------------------
//
// "techorate_borg.pov" -- Torsten Crass, 2008
//
// A demo application of "techorator.inc", a POV-Ray macro package
// for applying arbitrary itmes to arbitrary objects' surfaces in
// order to make them look "technical"
//
// This scene file demonstrates how to use the techorator for
// creating objects reminescent of the fameous Borg space ships
// from the Star Trek(R) TV series and motion pictures
//
//---------------------------------------------------------------------



// For demonstrating the techorator, we have to include it, ne c'est pas?
#include "techorator.inc"

// If you just wanna have a glance at the items used for techorating
// the Borg vessels, set this to 1:
#declare showItems = 0;



//--- Texture definitions ---------------------------------------------

// "Invisible" texture
#declare T_None = texture {
  pigment {color rgbt <1, 1, 1, 1>}
}

// Borg vessel texture
#declare T_Borg = texture {
  pigment {
    color rgb <0.3, 0.3, 0.3>
  }
  finish {
    ambient 0.1
    diffuse 0.7
    brilliance 0.5
    phong 1
    phong_size 10

  }
}

// Star surface
#declare T_StarSurface = texture {
  pigment {
    bozo
    color_map {
      [0.3 color rgb <1, 1, 0>]
      [0.5 color rgb <1, 0.6, 0>]
      [0.6 color rgb <1, 0.6, 0>]
      [0.8 color rgb <1, 1, 0>]
    }
  }
  finish {
    ambient 0.9
  }
  scale 20000/max(image_width, image_height)
}

// Sun spots
#declare T_SunSpots = texture {
  pigment {
    bozo
    color_map {
      [0.12 color rgbt <0.1, 0, 0, 0>]
      [0.22 color rgbt <1, 0.6, 0, 0>]
      [0.32 color rgbt <1, 0.6, 0, 1>]
    }
  }
  finish {
    ambient 0.9
  }
  scale 250
}

// Solar flare
#declare T_SolarFlare = texture {
  pigment {
    onion
    color_map {
      [0.4 color rgbt <1, 0.6, 0, 1>]
      [0.5 color rgbt <1, 0.6, 0, 0>]
      [0.6 color rgbt <1, 1, 0, 0>]
      [0.7 color rgbt <1, 1, 0, 0>]
      [0.8 color rgbt <1, 1, 0, 0>]
      [0.9 color rgbt <1, 0.6, 0, 1>]
    }
    turbulence 0.2
  }
  finish {
    ambient 1
  }
}

// An auto-scaling white-stars-on-black-background texture
#declare P_Space = pigment {
  bozo
  color_map {
    [0.8 color rgb 0]
    [0.85 color rgb 1]
  }
  scale 1.5/max(image_width, image_height)
}

// Green Borg glow
#declare C_BorgGreen = color rgb 1.5*<0.4, 1.0, 0>;



//--- Basic scene setup -----------------------------------------------

// Main illumination
light_source {
  <400000, 100000, -400000>
  color rgb 1.0
}

// Camera
camera {
  location  <0, 0, -35>
  look_at   <0, 0, 0>
  right     x*image_width/image_height
}

// The final frontier...
sky_sphere {
  pigment {P_Space}
}



//--- Macro definitions -----------------------------------------------

// Creates a thin, disk-like cylinder with a glowing texture of color col;
// glow intensity is i from center to r1 and linearly decreases from i at
// r1 down to 0 at r2. The cylinder's axis reaches from the origin to
// <0, -0.01, 0>.
#macro createGlow(r1, r2, col, i)
  disc {<0,0,0>, -0.01*z, r2
    texture {
      pigment {
        onion
        color_map {
          [r1/r2 color rgbt <col.red, col.green, col.blue, 1-i>]
          [1     color rgbt <col.red, col.green, col.blue, 1>]
        }
      }
      finish {
        ambient 1
        diffuse 0
      }
      scale r2
    }
    no_shadow
  }
#end



//--- Items for techorating Borg vessels ------------------------------

#declare item1 = union {
  union {
    cylinder {<-0.5, 0, 0> <-0.5, 0, 0.5> 0.05}
    cylinder {<-0.4, 0, -0.5> <-0.4, 0, 0.5> 0.05}
    cylinder {<-0.3, 0, 0> <-0.3, 0, 0.5> 0.05}
    cylinder {<-0.2, 0, 0> <-0.2, 0, 0.5> 0.05}
    cylinder {<-0.1, 0, 0> <-0.1, 0, 0.5> 0.05}
    box {<0.35, -0.2, -0.5> <0.45, 0.05, 0.5>}
    box {<-0.5, -0.2, 0.3> <0.5, 0.05, 0.4>}
  }
  intersection {
    box {<-0.5, -1, -0.1> <0.5, 1, -0.5>}
    cylinder {-0.5*z, 0, 3 translate -2.95*y}
  }
  texture {T_Borg}
}

#declare item2 = union {
  box {<-0.5, -0.1, -0.4> <0.5, 0.05, 0.4>}
  cylinder {<-0.4, 0, -0.5> <-0.4, 0, 0.5> 0.05}
  cylinder {<0.4, 0, -0.5> <0.4, 0, 0.5> 0.05}
  cylinder {0, 0.2*y, 0.3}
  torus {0.3, 0.025 translate 0.175*y}
  cylinder {-0.25*x, 0.25*x, 0.0125 rotate 20*y translate 0.2*y}
  cylinder {-0.25*x, 0.25*x, 0.0125 rotate 140*y translate 0.2*y}
  cylinder {-0.25*x, 0.25*x, 0.0125 rotate 260*y translate 0.2*y}
  texture {T_Borg}
}

#declare item3 = union {
  difference {
    box {<-0.5, -0.1, -0.4> <0.5, 0.05, 0.4>}
    union {
      box {<-0.3, -0.2, -0.3> <0.3, 0.1, 0>}
      plane {-x-z, -0.3}
    }
  }
  cylinder {<-0.4, 0, -0.5> <-0.4, 0, 0.5> 0.05}
  cylinder {<0.4, 0, -0.5> <0.4, 0, 0.5> 0.05}
  cylinder {<0.5, 0, 0.5> <0, 0, 0.0> 0.05}
  sphere {<-0.0, 0, 0.2> 0.1}
  sphere {<-0.2, 0, 0.2> 0.1}
  texture {T_Borg}
}

#declare item4 = union {
  difference {
    box {<-0.5, -0.1, -0.4> <0.5, 0.05, 0.4>}
    plane {-x-z, -0.3}
  }
  difference {
    union {
      sphere {<-0.05, 0, -0.05> 0.3}
      sphere {<-0.05, 0.3, -0.05> 0.025}
    }
    plane {y 0}
  }
  cylinder {<-0.4, 0, -0.5> <-0.4, 0, 0.5> 0.05}
  cylinder {<0.4, 0, -0.5> <0.4, 0, 0.5> 0.05}
  cylinder {<0.4, 0, 0.5> <0, 0, 0.1> 0.05}
  texture {T_Borg}
}

#declare item5 = union {
  difference {
    box {<-0.5, -0.1, -0.4> <0.5, 0.05, 0.4>}
    plane {-x-z, -0.3}
  }
  cylinder {<-0.4, 0, -0.5> <-0.4, 0, 0.5> 0.05}
  cylinder {<0.4, 0, -0.5> <0.4, 0, 0.5> 0.05}
  cylinder {<0.4, 0, 0.5> <0, 0, 0.1> 0.05}
  cylinder {<-0.3, -0.2, -0.3> <-0.3, 0.3, -0.3> 0.025}
  cylinder {<0.3, -0.2, -0.3> <0.3, 0.3, -0.3> 0.025}
  cylinder {<-0.3, 0.3, -0.3> <0.3, 0.3, -0.3> 0.025}
  sphere {<-0.3, 0.3, -0.3> 0.025}
  sphere {<0.3, 0.3, -0.3> 0.025}
  cylinder {<-0.3, 0, -0.1> <-0.3, 0, 0.3> 0.15}
  sphere {<-0.3, 0, -0.1> 0.15}
  sphere {<-0.3, 0, 0.3> 0.15}
  cylinder {<-0.0, 0, -0.1> <-0.0, 0, 0.3> 0.15}
  sphere {<-0.0, 0, -0.1> 0.15}
  sphere {<-0.0, 0, 0.3> 0.15}
  texture {T_Borg}
}


// Set up the array containing the techoration items
#declare items = array[5] {
  item1, item2, item3, item4, item5
}
// Determine number of items; prevents us from having to manually
// change the following array declarations
#local nItems = dimension_size(items, 1);


// Do you wanna get an overview over the declared items?
#if (showItems)

  // Get union object containing all items nicely aligned in a square grid
  object {displayTechorationItems(items)
    rotate 30*x
    translate <0, 1, -30.75>
  }


#else // No, thank you, just show me the techorated space ships

  // Scale all items to the minimum extend of their corresponding
  // grid node in A and B direction individually; makes them fit exactly
  #declare scaling = array[nItems] {
    <TECHORATE_SCALE_MIN, TECHORATE_SCALE_MIN, 0, 0>,
    <TECHORATE_SCALE_MIN, TECHORATE_SCALE_MIN, 0, 0>,
    <TECHORATE_SCALE_MIN, TECHORATE_SCALE_MIN, 0, 0>,
    <TECHORATE_SCALE_MIN, TECHORATE_SCALE_MIN, 0, 0>,
    <TECHORATE_SCALE_MIN, TECHORATE_SCALE_MIN, 0, 0>,
  }

  // Allow each item to be randomly rotated by 0, 90, 180, or 270
  // degrees; no further modification
  #declare modifications = array[nItems] {
    <0, 360, 4, 0, 0>,
    <0, 360, 4, 0, 0>,
    <0, 360, 4, 0, 0>,
    <0, 360, 4, 0, 0>,
    <0, 360, 4, 0, 0>,
  }

  // Pass items and what to do with them to the techorator
  setTechorationItems(items, scaling, modifications)

  // Tell that we want all items to be used with equal probabilities
  #declare weights = array[nItems] {
    1, 1, 1, 1, 1
  };
  setTechorationWeights(weights)


  //--- Building a Borg cube ------------------------------------------

  // Start off with an invisible, flat box which will serve a cube face
  #declare cube_face_shape = box {<-10, -10, -10> <10, 10, -9> texture {T_None}}

  // Set up a planar grid from <-10, -10, -11> to <10, 10, -11>, comprising 16
  // tiles in both A (= x) and B (= y) direction
  setTechorationDimensions(TECHORATE_PLANAR_GEOMETRY, -10, 10, 16, -10, 10, 16, 0, 0, 11)

  #declare cube_face = union {
    // Start techorating the cube face by placing 2 x 2 grid nodes sized items
    // up to a density of 0.2
    techorateInitialRandomGrid(cube_face_shape, 0.2, 2, 2)
    // Now fill up all remaining grid nodes with normal-size items
    finishGridTechoration(cube_face_shape)
  }

  // Build a cube from 6 appropriately rotated faces; some extra rotation
  // around z axis is added to prevent all faces from looking identical
  #declare elementary_cube = union {
    object {cube_face}
    object {cube_face rotate 90*z rotate 90*y}
    object {cube_face rotate 180*z rotate 180*y}
    object {cube_face rotate 270*z rotate 270*y}
    object {cube_face rotate 90*z rotate 90*x}
    object {cube_face rotate 270*z rotate -90*x}
  }

  // Place on such cube into a slightly bigger (and rotated) one to
  // add some visual "depth" to the whole thing
  #declare borg_cube = union {
    object {elementary_cube}
    object {elementary_cube rotate 90*y scale 1.1}
  }


  // We light each vessel individually to prevent green light leaking
  // through wholes in the structure to reach other objects in the scene
  light_group {

    // The borg cube
    object {borg_cube rotate 40*y}

    // The scary green glow from inside the cube
    light_source {0
      color C_BorgGreen
    }

    // Primary light source
    light_source {
      <400000, 100000, -400000>
      color rgb 1.0
    }

    // Some yellow-ish light as if coming from the star below
    light_source {
      <-10, -200, -20>
      color rgb 0.7*<1, 1, 0>
    }

    translate <10, 5, 15>
  }



  //--- Building a Borg sphere ------------------------------------------

  // Start of with an invisible sphere
  #declare base_sphere = sphere {0 10 texture {T_None}}

  // Set up a spherical grid, reaching all around the sphere's equator (longitude range
  // 0 - 360 degrees in 24 steps � 15 deg), and from -75 to 75 degrees latitude in 10
  // steps � 15 deg; scale grid by 20 in every direction
  setTechorationDimensions(TECHORATE_SPHERICAL_GEOMETRY, 0, 360, 24, -75, 75, 10, 20, 20, 20)

  #declare elementary_sphere = union {
    // Start techorating the sphere by placing 2 x 2 grid nodes sized items
    // up to a density of 0.1
    techorateInitialRandomGrid(base_sphere, 0.1, 2, 2)
    // Now fill up all remaining grid nodes with normal-size items
    finishGridTechoration(base_sphere)
  }

  // Build the final Borg sphere by placing differntly sized and rotated
  // spheres into each other
  #declare borg_sphere = union {
    object {elementary_sphere rotate 90*z scale 0.95}
    object {elementary_sphere scale 1.00}
    object {elementary_sphere rotate 90*x scale 1.05}
    // rotate the whole thing for increased visual appeal
    rotate 85*x
    rotate 7*y
  }

  // As mentioned already, we light each vessel individually
  light_group {

    // The Borg sphere
    object {borg_sphere scale 0.6}

    // The scary green glow from inside the sphere
    light_source {0
      color C_BorgGreen
    }

    // Primary light source
    light_source {
      <400000, 100000, -400000>
      color rgb 1.0
    }

    // Some yellow-ish light as if coming from the star below
    light_source {
      <-0, -200, -10>
      color rgb 0.7*<1, 1, 0>
    }

    translate <-6, -6, -5>
  }

#end




//--- Building a star -------------------------------------------------


// Initially, we declare the star as a large sphere...
#declare star = sphere {0 10000
  // ...with a hot surface...
  texture {T_StarSurface}
  // ...and some spots
  texture {T_SunSpots}
  no_shadow
}

// Random seed for scaling solar flares
#declare rs = seed(0);

// Re-definition of techorator's macro which normally yields an item
// from the items array according to the ttype parameter; we, however, want
// it to return *variable* items
#macro getTechorationItem(ttype, i, j, sVec, nVec)
  // Randomly choose a radius for the flare
  #local r = 500 + 1000 * rand(rs);
  // A very thin cylinder with radius s
  cylinder {-z, z, r
    // Texture is just a yellow-glowing ring
    texture {T_SolarFlare
      scale r
    }
    no_shadow
  }
#end

// The item list consists of a dummy item which will never be used
// (since we re-declared the getTechorationItem macro, remember?)
#declare items = array[1] {
  box {0, 0}
}
#declare nItems = dimension_size(items, 1);

// We won't use grid-based techoration, so this array contains
// another dummy entry
#declare scaling = array[nItems] {
  <0, 0, 0, 0>
}

// We will, however, make use of random rotation
#declare modifications = array[nItems] {
  <0, 360, 360, 0, 0>
}

// Pass (dummy) item and what to do with it to the techorator
setTechorationItems(items, scaling, modifications)

// Array with dummy entry for the (only) item's frequency
#declare weights = array[nItems] {
  1
}
setTechorationWeights(weights)

// Set up a sphere (even larger than the star) as hull shape on which
// to randomly select points defining where to place items
setTechorationDimensions(TECHORATE_SPHERICAL_GEOMETRY, 0, 360, 0, -90, 90, 0, 20000, 20000, 20000)

// Coose a different random seed for the techorator (took some experiments to
// find a seed value leading to a good-looking distribution of flares)
setTechoratorRandomSeed(41)

// Randomly place 60 solar flares on the star's surface
#declare star = union {
  object {star}
  techorateRandomly(star, 60)
}


// Place star in background
object {star translate <0, -10000, 10000>}

// Create the illusion of a glowing star atmosphere
object {createGlow(10000, 15000, <1, 1, 0.5>, 0.5) rotate -5*x translate <0, -10000, 10000>}

